@@ -5,20 +5,17 @@ use Moose;
 
 use Carp qw(croak);
 
+use File::NFSLock;
+use IO::AtomicFile;
 use JSON;
 
-use File::Spec;
-use File::Path qw(remove_tree make_path);
-
-use Directory::Transactional;
-
-use Data::Stream::Bulk::Util qw(bulk);
+use Data::Stream::Bulk::Path::Class;
 
 use MooseX::Types::Path::Class qw(Dir File);
 
 use namespace::clean -except => 'meta';
 
-our $VERSION = "0.02_02";
+our $VERSION = "0.01";
 
 with qw(
     KiokuDB::Backend
@@ -26,8 +23,6 @@ with qw(
     KiokuDB::Backend::Role::Clear
     KiokuDB::Backend::Role::Scan
     KiokuDB::Backend::Role::Query::Simple::Linear
-    KiokuDB::Backend::Role::TXN
-    KiokuDB::Backend::Role::TXN::Nested
 );
 
 sub BUILD {
@@ -55,17 +50,17 @@ has create => (
 );
 
 has object_dir => (
-    isa => "Str",
+    isa => Dir,
     is  => "ro",
-    default => "all",
+    lazy_build => 1,
 );
 
-has root_set_dir => (
-    isa => "Str",
-    is  => "ro",
-    default => "root",
-);
+sub _build_object_dir {
+    my $self = shift;
+    $self->dir->subdir("all");
+}
 
+# TODO implement trie fanning on disk
 has trie => (
     isa => "Bool",
     is  => "ro",
@@ -86,23 +81,41 @@ has trie_levels => (
     default => 2,
 );
 
-has _txn_manager => (
-    isa => "Directory::Transactional",
+has root_set_dir => (
+    isa => Dir,
     is  => "ro",
     lazy_build => 1,
 );
 
-sub _build__txn_manager {
+sub _build_root_set_dir {
     my $self = shift;
-
-    Directory::Transactional->new(
-        root => $self->dir,
-    );
+    $self->dir->subdir("root");
 }
 
-sub txn_begin { shift->_txn_manager->txn_begin(@_) }
-sub txn_commit { shift->_txn_manager->txn_commit(@_) }
-sub txn_rollback { shift->_txn_manager->txn_rollback(@_) }
+has lock => (
+    isa => "Bool",
+    is  => "rw",
+    default => 1,
+);
+
+has lock_file => (
+    isa => File,
+    is  => "ro",
+    lazy_build => 1,
+);
+
+sub _build_lock_file {
+    my $self = shift;
+    $self->dir->file("lock");
+}
+
+sub write_lock {
+    my $self = shift;
+
+    return 1 unless $self->lock;
+
+    File::NFSLock->new({ file => $self->lock_file->stringify, lock_type => "EXCLUSIVE" });
+}
 
 sub get {
     my ( $self, @uids ) = @_;
@@ -123,11 +136,9 @@ sub delete {
 
     my @uids = map { ref($_) ? $_->id : $_ } @ids_or_entries;
 
-    my $t = $self->_txn_manager;
-
     foreach my $uid ( @uids ) {
         foreach my $file ( $self->object_file($uid), $self->root_set_file($uid) ) {
-            $t->unlink($file);
+            $file->remove;
         }
     }
 }
@@ -135,8 +146,7 @@ sub delete {
 sub exists {
     my ( $self, @uids ) = @_;
 
-    my $t = $self->_txn_manager;
-    map { $t->exists($self->object_file($_)) } @uids;
+    map { -e $self->object_file($_) } @uids;
 }
 
 sub get_entry {
@@ -150,7 +160,7 @@ sub get_entry {
 sub open_entry {
     my ( $self, $id ) = @_;
 
-    $self->_txn_manager->openr( $self->object_file($id) );
+    $self->object_file($id)->openr;
 }
 
 sub insert_entry {
@@ -160,73 +170,70 @@ sub insert_entry {
 
     my $file = $self->object_file($id);
 
-    my $t = $self->_txn_manager;
+    $file->parent->mkpath unless -d $file->parent;
 
-    my $fh = $t->openw($file);
+    my $fh = IO::AtomicFile->open( $file, "w" );
 
     $self->serializer->serialize_to_stream($fh, $entry);
 
-    close $fh || croak "Couldn't store: $!";
+    {
+        my $lock = $self->write_lock;
 
-    my $root_file = $self->root_set_file($id);
+        $fh->close || croak "Couldn't store: $!";
 
-    if ( $entry->root ) {
-        # just create a file with the same name... this used to be a hardlink
-        # but that doesn't play nice with Directory::Transactional
-        # maybe it will return eventually
-        $t->openw($root_file);
-    } else {
-        $t->unlink($root_file);
+        my $root_file = $self->root_set_file($id);
+        $root_file->remove;
+
+        if ( $entry->root ) {
+            $root_file->parent->mkpath unless -d $root_file->parent;
+            link( $file, $root_file );
+        }
     }
 }
 
-sub _trie_path {
-    my ( $self, @path ) = @_;
+sub _trie_dir {
+    my ( $self, $dir, $uid ) = @_;
 
-    return File::Spec->catfile(@path) unless $self->trie;
-
-    my $uid = pop @path;
+    return $dir unless $self->trie;
 
     my $id_hex = unpack("H*", $uid);
 
     my $nyb = $self->trie_nybbles;
 
     for ( 1 .. $self->trie_levels ) {
-        push @path, substr($id_hex, 0, $nyb, '');
+        $dir = $dir->subdir( substr($id_hex, 0, $nyb, '') );
     }
 
-    File::Spec->catfile( @path, $uid );
+    return $dir;
 }
 
 sub object_file {
     my ( $self, $uid ) = @_;
 
-    $self->_trie_path( $self->object_dir, $uid);
+    my $dir = $self->_trie_dir( $self->object_dir, $uid);
+
+    $dir->file($uid);
 }
 
 sub root_set_file {
     my ( $self, $uid ) = @_;
 
-    $self->_trie_path( $self->root_set_dir, $uid );
+    my $dir = $self->_trie_dir( $self->root_set_dir, $uid);
+
+    $dir->file($uid);
 }
 
 sub create_dirs {
     my $self = shift;
 
-    make_path( $self->object_dir );
-    make_path( $self->root_set_dir );
+    $self->object_dir->mkpath;
+    $self->root_set_dir->mkpath;
 }
 
 sub clear {
     my $self = shift;
 
-    my $m = $self->_txn_manager;
-
-    my $stream = $m->file_stream( only_files => 1 );
-
-    while ( my $block = $stream->next ) {
-        $m->unlink($_) for @$block;
-    }
+    $_->rmtree({ keep_root => 1 }) for $self->root_set_dir, $self->object_dir;
 }
 
 sub all_entries {
@@ -234,31 +241,19 @@ sub all_entries {
 
     my $ser = $self->serializer;
 
-    my $t = $self->_txn_manager;
-
-    my $stream = $t->file_stream( only_files => 1, dir => $self->object_dir );
-
-    $stream->filter(sub { [ map { $ser->deserialize_from_stream($t->openr($_)) } @$_ ]});
+    Data::Stream::Bulk::Path::Class->new( dir => $self->object_dir, only_files => 1 )->filter(sub { [ map {
+        $ser->deserialize_from_stream( $_->openr );
+    } @$_ ]});
 }
 
-# broken:
 sub root_entries {
     my $self = shift;
 
     my $ser = $self->serializer;
 
-    my $t = $self->_txn_manager;
-
-    my $stream = $t->file_stream( only_files => 1, dir => $self->root_set_dir );
-
-    my $ids = $stream->filter(sub {[
-        map {
-            my ( undef, undef, $file ) = File::Spec->splitpath($_);
-            $file;
-        } @$_
-    ]});
-
-    $ids->filter(sub{[ $self->get(@$_) ]});
+    Data::Stream::Bulk::Path::Class->new( dir => $self->root_set_dir, only_files => 1 )->filter(sub { [ map {
+        $ser->deserialize_from_stream( $_->openr );
+    } @$_ ]});
 }
 
 __PACKAGE__->meta->make_immutable;
@@ -282,8 +277,8 @@ KiokuDB::Backend::Files - One file per object backend
 
 =head1 DESCRIPTION
 
-This backend provides a file based backend using L<Directory::Transactional> to
-provide ACID semantics.
+This backend provides a file based backend, utilizing L<IO::AtomicFile> and
+L<File::NFSLock> for safety.
 
 This is one of the slower backends, and the support for searching is very
 limited (only a linear scan is supported), but it is suitable for small, simple
@@ -301,15 +296,21 @@ The directory for the backend.
 
 If true (defaults to false) the directories will be created at instantiation time.
 
+=item lock
+
+Whether or not locking is enabled.
+
+Defaults to true.
+
 =item object_dir
 
-Defaults to C<all>.
+Defaults to the subdirectory C<all> of C<dir>
 
-=item root_set_dir
+=item root_dir
 
-Defaults C<root>.
+Defaults to the subdirectory C<root> of C<dir>
 
-Root set entries are symlinked into this directory as well.
+Root set entries are hard linked into this directory as well.
 
 =item trie
 
